/*
** Copyright (C) 2009 Patrick Kirchhof, Kevin Suske and Pasqual Dhring
** 
** This file is part of GFA.
** 
** GFA is free software: you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation, either version 3 of the License, or
** (at your option) any later version.
** 
** GFA is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with GFA.  If not, see <http://www.gnu.org/licenses/>.
** 
*/

package animator.src;

import java.io.File;

import java.text.SimpleDateFormat;

import java.util.Calendar;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import org.w3c.dom.Element;

public class AnimatorBuilder {

	// used to keep track of all AnimatorObjects
	public static HashMap<String, AnimatorObject> objects = new HashMap<String, AnimatorObject>();
	// used to keep track of the attributes of the structure currently being processed
	private HashMap<String, String> nodeAttributes = new HashMap<String, String>();
	private String firstLevelBranch = "", secondLevelBranch = "", currentTagValue = "";
	private String touchsensorName = "TOUCHSENSOR_StartAnimation", timesensorName = "TIME";
	private String simulationLength = "";
	private Double xmin, xmax, ymin, ymax;
	private Element head, scene;
	private AnimatorX3D ax3d;
	
	public AnimatorBuilder() {		
		ax3d = new AnimatorX3D();
		head = ax3d.createNodeHead();
		scene = ax3d.createNodeScene();
		
		// add metadata
		head.appendChild(ax3d.createNodeMeta("title", AnimatorUtils.x3dFileName+".x3d"));
		head.appendChild(ax3d.createNodeMeta("created", new SimpleDateFormat("MM/dd/yyyy").format(Calendar.getInstance().getTime())));
		head.appendChild(ax3d.createNodeMeta("generator", "genericFactoryAnimator3D"));
		
		// add standard proto declarations with (alternate) locations
		scene.appendChild(ax3d.createNodeExternProtoDeclare("StandardWorkstation", new String[]{
				AnimatorUtils.templateFolder+"Proto_StandardWorkstation.x3d",
				".\\genericFactoryAnimator3D\\templates\\Proto_StandardWorkstation.x3d",
				AnimatorUtils.templateFolderRel+"Proto_StandardWorkstation.x3d",
				"Proto_StandardWorkstation.x3d"}));
		scene.appendChild(ax3d.createNodeExternProtoDeclare("StandardJob", new String[]{
				AnimatorUtils.templateFolder+"Proto_StandardJob.x3d", 
				"Proto_StandardJob.x3d"}));
		scene.appendChild(ax3d.createNodeExternProtoDeclare("StandardTransporter", new String[]{
				AnimatorUtils.templateFolder+"Proto_StandardTransporter.x3d", 
				"Proto_StandardTransporter.x3d"}));
		
		// add user-defined proto declarations (if set) with (alternate) locations
		Iterator it = AnimatorBuilder.objects.entrySet().iterator();
    	while (it.hasNext()) { 
    		Map.Entry entry = (Map.Entry) it.next();
    		AnimatorObject ao = (AnimatorObject)(entry).getValue();
    		
    		if (AnimatorUtils.isProtoDeclaration(ao.getTemplateURL()))
    			scene.appendChild(ax3d.createNodeExternProtoDeclare(AnimatorUtils.getProtoName(ao.getTemplateURL()), new String[]{
    				ao.getTemplateURL(), 
    				"..\\..\\"+new File(ao.getTemplateURL()).getName(),
    				"..\\"+new File(ao.getTemplateURL()).getName(),
    				new File(ao.getTemplateURL()).getName()}));
    		
    		ao.resetObject();
    	}
		
		// user interface
		if (AnimatorUtils.useExtensiveMenu) {
			// use extensive x3d user interface (only compatible with octaga) + (alternate) locations
			scene.appendChild(ax3d.createNodeExternProtoDeclare("TimeMenu", new String[]{
					AnimatorUtils.menuFolder+"TimeMenuPrototype.x3d",
					AnimatorUtils.menuFolderRel+"TimeMenuPrototype.x3d",
					"..\\menu\\TimeMenuPrototype.x3d",
					"..\\..\\TimeMenuPrototype.x3d",
					"..\\TimeMenuPrototype.x3d",
					"TimeMenuPrototype.x3d"}));
		} else {
			// use simple x3d user interaface (start button only, more compatible)
			scene.appendChild(ax3d.createCustomGroupUI(touchsensorName));
			scene.appendChild(ax3d.createNodeRoute("ProximitySensorOrigin", "position_changed", "UserInterface", "set_translation"));
			scene.appendChild(ax3d.createNodeRoute("ProximitySensorOrigin", "orientation_changed", "UserInterface", "set_rotation"));
		}
	}
	
	/**
	 * Add an event to the event list of a job object (or transporter object)
	 */
	private void addEvent() {
		simulationLength = nodeAttributes.get("time");
		String uuid = nodeAttributes.get("object"); String defID = AnimatorUtils.jobPrefix + uuid;
		String action = nodeAttributes.get("action");
		
		if (action.equals("addJob")) createEventAddJob(uuid, defID, simulationLength);
		if (action.equals("changeState")) createEventChangeState(uuid, defID, simulationLength);
		if (action.equals("create")) createEventCreate(uuid, defID, simulationLength);
		if (action.equals("insert")) createEventInsert(uuid, defID, simulationLength);
		if (action.equals("loadTransporter")) createEventLoadTransporter(uuid, simulationLength);
		if (action.equals("moveJob")) createEventMoveJob(uuid, simulationLength);
		if (action.equals("moveTransporter")) createEventMoveTransporter(uuid, simulationLength);
		if (action.equals("remove")) createEventRemove(uuid, defID, simulationLength);
		if (action.equals("seize")) createEventSeize(uuid, defID, simulationLength);		
		if (action.equals("terminate")) createEventTerminate(uuid, defID, simulationLength);
		if (action.equals("unloadTransporter")) objects.get(uuid).setTransported(false, nodeAttributes.get("transporter"));
	}
	
	/**
	 * Add an object to the x3d output
	 */
	private void addObject() {
		String uuid = nodeAttributes.get("uuid");
		String name = nodeAttributes.get("name");
		String type = nodeAttributes.get("type");
		
		if (type.equals("conveyor")) createObjectConveyor(uuid, name);
		if (type.equals("parking")) createObjectParking(uuid, name);
		if (type.equals("path")) createObjectPath(uuid, name);
		if (type.equals("queue")) createObjectQueue(uuid, name);
		if (type.equals("sink")) createObjectSink(uuid, name);
		if (type.equals("source")) createObjectSource(uuid, name);
		if (type.equals("transporter")) createObjectTransporter(uuid, name);
		if (type.equals("workstation")) createObjectWorkstation(uuid, name);
	}
	
	/**
	 * Create the animation (visibility, position) for all jobs and write the information to x3d
	 */
	private void createAnimationJobs() {
		Iterator it = objects.entrySet().iterator();
		String def, vkey, vkeyValue, pkey, pkeyValue;
		while (it.hasNext()) {
			AnimatorObject ao = (AnimatorObject)((Map.Entry)it.next()).getValue();
			
			if (ao.getType().equals("job") && ao.getEventCount() > 0) {				
				def = ao.getDefID(); vkey = "0"; vkeyValue="0"; pkey = "0"; pkeyValue = "";				

				// add job object
				createTemplate(ao, def+"trans", def, "StandardJob", ao.getPosition(false, true, true), ao.getColorX3D());
										    	
		    	for (int i=0; i < ao.getEventCount(); i++) {
		    		// visibility
		    		vkey = vkey + " " + AnimatorUtils.getTimeKey(ao.getEventTime(i), simulationLength);
		    		if (ao.getEventStatus(i).equals("true")) vkeyValue = vkeyValue + " 1";
		    		if (ao.getEventStatus(i).equals("false")) vkeyValue = vkeyValue + " 0";
		    		// position
		    		if (i == 0) pkeyValue = ao.getEventPosition(i) + " "; // time 0
		    		pkey = pkey + " " + AnimatorUtils.getTimeKey(ao.getEventTime(i), simulationLength);
		    		pkeyValue = pkeyValue + ao.getEventPosition(i);
		    		if (i != ao.getEventCount()-1) pkeyValue = pkeyValue + " ";
		    	}
		    	scene.appendChild(ax3d.createNodeIntegerSequencer(def+"_sequencer", vkey, vkeyValue));
		    	scene.appendChild(ax3d.createNodeRoute(timesensorName, "fraction_changed", def+"_sequencer", "set_fraction"));
		    	scene.appendChild(ax3d.createNodeRoute(def+"_sequencer", "value_changed", def, "setVisibilityInt"));
		    			    	
		    	scene.appendChild(ax3d.createNodePositionInterpolator(def+"pos", pkey, pkeyValue));
		    	scene.appendChild(ax3d.createNodeRoute(timesensorName, "fraction_changed", def+"pos", "set_fraction"));
		    	scene.appendChild(ax3d.createNodeRoute(def+"pos", "value_changed", def+"trans", "translation"));
			}
		}	
	}
	
	/**
	 * Create the animation (load status, position, orientation) for all transporters 
	 * and write the information to x3d
	 */
	private void createAnimationTransporters() {
		Iterator it = objects.entrySet().iterator();
		String def, lkey, lkeyValue, pkey, pkeyValue, okey, okeyValue, okeyValuePrev = " 0 1 0 0";
		double angle = 0, rotation, timeKeyTmp;
		while (it.hasNext()) {
			AnimatorObject ao = (AnimatorObject)((Map.Entry)it.next()).getValue();
			
			if (ao.getType().equals("transporter") && ao.getEventCount() > 0) {				
				def = ao.getDefID(); lkey = "0"; lkeyValue="0"; pkey = "0"; pkeyValue = ""; okey = "0"; okeyValue = "0 1 0 0";
		    	
		    	for (int i=0; i < ao.getEventCount(); i++) {
		    		// load
		    		lkey = lkey + " " + AnimatorUtils.getTimeKey(ao.getEventTime(i), simulationLength);
		    		lkeyValue = lkeyValue + " " + ao.getEventStatus(i);
		    		// position
		    		if (i == 0) pkeyValue = ao.getEventPosition(i) + " "; // time 0
		    		pkey = pkey + " " + AnimatorUtils.getTimeKey(ao.getEventTime(i), simulationLength);
		    		pkeyValue = pkeyValue + ao.getEventPosition(i);
		    		if (i != ao.getEventCount()-1) pkeyValue = pkeyValue + " ";
		    		// orientation		    		
		    		if (i < ao.getEventCount()-1) 
		    			angle = AnimatorUtils.getAngle(Double.parseDouble(AnimatorUtils.getX(ao.getEventPosition(i))), 
		    					Double.parseDouble(AnimatorUtils.getX(ao.getEventPosition(i+1))), 
		    					Double.parseDouble(AnimatorUtils.getY(ao.getEventPosition(i))), 
		    					Double.parseDouble(AnimatorUtils.getY(ao.getEventPosition(i+1))));
		    		timeKeyTmp = Double.parseDouble(AnimatorUtils.getTimeKey(ao.getEventTime(i), simulationLength)) - 0.0001;
		    		okey = okey + " " + String.valueOf(timeKeyTmp) + " " + AnimatorUtils.getTimeKey(ao.getEventTime(i), simulationLength);
		    		if (angle == -45) rotation = Math.toRadians(angle) - Math.toRadians(90);
		    			else rotation = Math.toRadians(180) - Math.toRadians(angle);
		    		okeyValue = okeyValue + okeyValuePrev + " 0 1 0 " + rotation;
		    		okeyValuePrev = " 0 1 0 "+rotation;
		    	}
		    	
		    	scene.appendChild(ax3d.createNodeIntegerSequencer(def+"_sequencer", lkey, lkeyValue));
		    	scene.appendChild(ax3d.createNodeRoute(timesensorName, "fraction_changed", def+"_sequencer", "set_fraction"));
		    	scene.appendChild(ax3d.createNodeRoute(def+"_sequencer", "value_changed", def, "seize"));
		    			    	
		    	scene.appendChild(ax3d.createNodePositionInterpolator(def+"pos", pkey, pkeyValue));
		    	scene.appendChild(ax3d.createNodeRoute(timesensorName, "fraction_changed", def+"pos", "set_fraction"));
		    	scene.appendChild(ax3d.createNodeRoute(def+"pos", "value_changed", def+"trans", "translation"));
		    	
		    	scene.appendChild(ax3d.createNodeOrientationInterpolator(def+"orient", okey, okeyValue));
		    	scene.appendChild(ax3d.createNodeRoute(timesensorName, "fraction_changed", def+"orient", "set_fraction"));
		    	scene.appendChild(ax3d.createNodeRoute(def+"orient", "value_changed", def+"trans", "rotation"));
			}
		}	
	}
	
	/**
	 * Create the animation (status changes) for all workstations and write the information to x3d
	 */
	private void createAnimationWorkstations() {
		Iterator it = objects.entrySet().iterator();
		String def, key, keyValue;
		while (it.hasNext()) {
			AnimatorObject ao = (AnimatorObject)((Map.Entry)it.next()).getValue();
			
			if (ao.getType().equals("workstation") && ao.getEventCount() > 0) {
				def = ""; key = "0"; keyValue="0";

		    	for (int i=0; i < ao.getEventCount(); i++) {
		    		def = ao.getDefID();	
		    		key = key + " " + AnimatorUtils.getTimeKey(ao.getEventTime(i), simulationLength);
		    		if (ao.getEventStatus(i).equals("idle")) keyValue = keyValue + " 0";
		    		if (ao.getEventStatus(i).equals("busy")) keyValue = keyValue + " 1";
		    		if (ao.getEventStatus(i).equals("breakdown")) keyValue = keyValue + " 2";
		    		if (ao.getEventStatus(i).equals("stopped")) keyValue = keyValue + " 3";
		    		if (ao.getEventStatus(i).equals("maintained")) keyValue = keyValue + " 4";
		    		if (ao.getEventStatus(i).equals("setup")) keyValue = keyValue + " 5";
		    	}
		    	
		    	scene.appendChild(ax3d.createNodeIntegerSequencer(def+"_sequencer", key, keyValue));
		    	scene.appendChild(ax3d.createNodeRoute(timesensorName, "fraction_changed", def+"_sequencer", "set_fraction"));
		    	scene.appendChild(ax3d.createNodeRoute(def+"_sequencer", "value_changed", def, "setStatus"));
			}	
		}
	}
	
	/**
	 * Create a baseplate for the objects - this can be a solid one, a grid or both
	 */
	private void createBaseplate(boolean grid, boolean solid, double gridz, double solidz, double margin) {
		double xminm = xmin-margin, yminm = ymin-margin, xmaxm = xmax+margin, ymaxm = ymax+margin;
		if (grid == true) {
			for (double i=xminm; i <= xmaxm; i++) {
				scene.appendChild(ax3d.createBranchLineSet("", "2", 
						i+" "+gridz+" -"+yminm+" "+i+" "+gridz+" -"+ymaxm, ".8", "1 1 1", "0"));
			}
			for (double j=yminm; j <= ymaxm; j++) {
				scene.appendChild(ax3d.createBranchLineSet("", "2", 
						xminm+" "+gridz+" -"+j+" "+xmaxm+" "+gridz+" -"+j, ".8", "1 1 1", "0"));
			}
		}
		
		if (solid == true) {
			scene.appendChild(ax3d.createBranchQuadSet("", 
					xminm+" "+solidz+" -"+yminm+" "+
					xmaxm+" "+solidz+" -"+yminm+" "+
					xmaxm+" "+solidz+" -"+ymaxm+" "+
					xminm+" "+solidz+" -"+ymaxm, "", ".15 .15 .15"));
		}
	}
	
	/**
	 * Adds a job to the model 
	 */
	private void createEventAddJob(String uuid, String defID, String currentTime) {
		AnimatorObject job = new AnimatorObject(defID, uuid, "", "job");
		String item = nodeAttributes.get("item");
		
		// get template url and color by item uuid
		Iterator it = objects.entrySet().iterator();
		while (it.hasNext()) {
			AnimatorObject ao = (AnimatorObject)((Map.Entry)it.next()).getValue();
			if (ao.getType().equals("item") && ao.getUUID().equals(item)) {
				job.setTemplateURL(ao.getTemplateURL());
				job.setColor(ao.getColor());
			}		
		}
		
		objects.put(uuid, job);
	}
	
	/**
	 * Add changeState event to the event list of an object (by uuid) 
	 */
	private void createEventChangeState(String uuid, String defID, String currentTime) {
		String state = nodeAttributes.get("state");
		objects.get(uuid).addEvent(currentTime, state);
	}
	
	/**
	 * Add create event to the event list of an object (by uuid) 
	 */
	private void createEventCreate(String uuid, String defID, String currentTime) {
		String source = nodeAttributes.get("source");
		// visibility
		String visibility = "true"; if (AnimatorUtils.enableTransporters) visibility = "false";
		
		AnimatorObject ao = objects.get(uuid);
		ao.addEvent(currentTime, visibility, objects.get(source).getPosition(false, true, true), source);
		ao.setPositionByString(objects.get(source).getPosition(false, true, true), true);
	}
	
	/**
	 * Add insert event to the event list of an object (by uuid) 
	 */
	private void createEventInsert(String uuid, String defID, String currentTime) {
		AnimatorObject job = objects.get(uuid);
		String queueUUID = nodeAttributes.get("queue");
		
		if (!job.getTransporterUUID().isEmpty())
			createEventUnloadTransporter(uuid, job.getTransporterUUID(), queueUUID, currentTime);
		
		job.addEvent(currentTime, "true", objects.get(queueUUID).getPosition(true, true, true), queueUUID);
	}
	
	/**
	 * Add loadtransporter event to the event list of an object (by transporteruuid) 
	 */
	private void createEventLoadTransporter(String uuid, String currentTime) {
		AnimatorObject job = objects.get(uuid);
		AnimatorObject transporter = objects.get(nodeAttributes.get("transporter"));
		String loadUUID = job.getEventUUID(job.getEventCount()-1);
		int currentLoad = Integer.parseInt(transporter.getEventStatus(transporter.getEventCount()-1));
		job.setTransported(true, "");
		
		if (transporter.getEventCount() == 0) createEventParking(transporter);
		
		// hold job position until job is loaded
		String visibility = "true"; if (AnimatorUtils.enableTransporters) visibility = "false";
		job.addEvent(currentTime, visibility, job.getEventPosition(job.getEventCount()-1), transporter.getUUID());
		
		transporter.addEvent(currentTime, String.valueOf(currentLoad+1), transporter.getEventPosition(transporter.getEventCount()-1), loadUUID);
	}
	
	/**
	 * Add movetransporter event to the event list of an object (by transporteruuid) 
	 */
	private void createEventMoveJob(String jobUUID, String currentTime) {
		AnimatorObject job = objects.get(jobUUID);	
		String posx = nodeAttributes.get("xposition"); String posy = nodeAttributes.get("yposition"); String posz = nodeAttributes.get("zposition");
		// visibility
		String visibility = "true"; if ((AnimatorUtils.enableTransporters) && (job.getTransported())) visibility = "false";

		job.addEvent(currentTime, visibility, posx+" "+posz+" -"+posy, "");
	}
	
	/**
	 * Add movetransporter event to the event list of an object (by transporteruuid) 
	 */
	private void createEventMoveTransporter(String transporterUUID, String currentTime) {
		AnimatorObject transporter = objects.get(transporterUUID);	
		String posx = nodeAttributes.get("xposition"); String posy = nodeAttributes.get("yposition"); String posz = nodeAttributes.get("zposition");
		
		if (transporter.getEventCount() == 0) createEventParking(transporter);
		
		int currentLoad = Integer.parseInt(transporter.getEventStatus(transporter.getEventCount()-1));		
		transporter.addEvent(currentTime, String.valueOf(currentLoad), posx+" "+posz+" -"+posy, "");
	}
	
	/**
	 * Add parking event to the top of the event list of a transporter
	 */
	private void createEventParking(AnimatorObject transporter) {
		// get uuid for parking and write it to the transporter event list
		Iterator it = objects.entrySet().iterator(); String parkingUUID = "";
		while (it.hasNext()) {
			AnimatorObject pg = (AnimatorObject)((Map.Entry)it.next()).getValue();
			if (pg.getType().equals("parking") && pg.getPosition(false, false, false).equals(transporter.getPosition(false, false, false)))
				parkingUUID = pg.getUUID();
		}			
		transporter.addEvent("0", "0", transporter.getPosition(false, false, true), parkingUUID);
	}
	
	/**
	 * Add remove event to the event list of an object (by uuid) 
	 */
	private void createEventRemove(String uuid, String defID, String currentTime) {
		String queueUUID = nodeAttributes.get("queue");
		// visibility
		String visibility = "true"; if (objects.get(uuid).getTransported()) visibility = "false";
		
		objects.get(uuid).addEvent(currentTime, visibility, objects.get(queueUUID).getPosition(false, true, true), queueUUID);
	}
	
	/**
	 * Add seize event to the event list of an object (by uuid) 
	 */
	private void createEventSeize(String uuid, String defID, String currentTime) {
		String duration = nodeAttributes.get("duration");
		String jobUUID = nodeAttributes.get("job");
		// visibility
		String visibility = "true"; if (objects.get(jobUUID).getTransported()) visibility = "false";
			
		objects.get(jobUUID).addEvent(currentTime, visibility, objects.get(uuid).getPosition(false, true, true), uuid);
		double untilTime = Double.parseDouble(currentTime) + Double.parseDouble(duration);
		objects.get(jobUUID).addEvent(String.valueOf(untilTime), visibility, objects.get(uuid).getPosition(false, true, true), uuid);
	}
	
	/**
	 * Add terminate event to the event list of an object (by uuid) 
	 */
	private void createEventTerminate(String uuid, String defID, String currentTime) {
		AnimatorObject job = objects.get(uuid);
		String sinkUUID = nodeAttributes.get("sink");
		// visibility (reset visibility)
		if (!job.getTransporterUUID().isEmpty())
			createEventUnloadTransporter(uuid, job.getTransporterUUID(), sinkUUID, currentTime);
		
		objects.get(uuid).addEvent(currentTime, "false", objects.get(sinkUUID).getPosition(false, true, true), sinkUUID);
	}
	
	/**
	 * Add unloadtransporter event to the event list of an object (by transporteruuid) 
	 */
	private void createEventUnloadTransporter(String uuid, String transporterUUID, String unloadUUID, String currentTime) {
		AnimatorObject job = objects.get(uuid);
		AnimatorObject transporter = objects.get(transporterUUID);
		
		transporter.addEvent(currentTime, "0", transporter.getEventPosition(transporter.getEventCount()-1), unloadUUID);
		job.setTransported(false, "");
	}
	
	/**
	 * Create a conveyor in x3d
	 */
	private void createObjectConveyor(String uuid, String name) {
		String defID = AnimatorUtils.conveyorPrefix + uuid;
		String xfrom = nodeAttributes.get("xfrom"), yfrom = nodeAttributes.get("yfrom"), zfrom = nodeAttributes.get("zfrom"), 
			xto = nodeAttributes.get("xto"), yto = nodeAttributes.get("yto"), zto = nodeAttributes.get("zto");
		
		scene.appendChild(ax3d.createBranchLineSet(defID, "2", 
				xfrom+" "+zfrom+" -"+yfrom+" "+xto+" "+zto+" -"+yto, ".6", "0 1 0", "2"));
		
		AnimatorObject ao = new AnimatorObject(defID, uuid, name, "conveyor");
		ao.setPosition(xfrom, yfrom, zfrom, xto, yto, zto);
		ao.setStdTravelTime(nodeAttributes.get("stdtraveltime"));
		objects.put(uuid, ao);
	}
	
	/**
	 * Create parking area in x3d - since the parking area is just one x-y-z coordinate,
	 * the actual x3d parking area is extended in both x and y directions by 1
	 */
	private void createObjectParking(String uuid, String name) {
		String defID = AnimatorUtils.parkingPrefix + uuid;
		double xminp = Double.parseDouble(nodeAttributes.get("xposition"))-1,
			xmaxp = Double.parseDouble(nodeAttributes.get("xposition"))+1,
			yminp = Double.parseDouble(nodeAttributes.get("yposition"))-1,
			ymaxp = Double.parseDouble(nodeAttributes.get("yposition"))+1;
		
		scene.appendChild(ax3d.createBranchQuadSet(defID, 
				xminp+" "+nodeAttributes.get("zposition")+" -"+yminp+" "+
				xmaxp+" "+nodeAttributes.get("zposition")+" -"+yminp+" "+
				xmaxp+" "+nodeAttributes.get("zposition")+" -"+ymaxp+" "+
				xminp+" "+nodeAttributes.get("zposition")+" -"+ymaxp, "", ".5 .5 .5"));
		
		AnimatorObject ao = new AnimatorObject(defID, uuid, name, "parking");
		ao.setPosition(nodeAttributes.get("xposition"), nodeAttributes.get("yposition"), nodeAttributes.get("zposition"));
		objects.put(uuid, ao);
	}
	
	/**
	 * Create a path in x3d
	 */
	private void createObjectPath(String uuid, String name) {
		String defID = AnimatorUtils.pathPrefix + uuid;
		String xfrom = nodeAttributes.get("xfrom"), yfrom = nodeAttributes.get("yfrom"), zfrom = nodeAttributes.get("zfrom"), 
			xto = nodeAttributes.get("xto"), yto = nodeAttributes.get("yto"), zto = nodeAttributes.get("zto");
		
		scene.appendChild(ax3d.createBranchLineSet(defID, "2", 
				xfrom+" "+zfrom+" -"+yfrom+" "+xto+" "+zto+" -"+yto, ".6", "0 0 1", "0"));
		
		AnimatorObject ao = new AnimatorObject(defID, uuid, name, "path");
		ao.setPosition(xfrom, yfrom, zfrom, xto, yto, zto);
		ao.setStdTravelTime(nodeAttributes.get("stdtraveltime"));
		objects.put(uuid, ao);
	}
	
	/**
	 * Create a queue in x3d
	 */
	private void createObjectQueue(String uuid, String name) {
		String defID = AnimatorUtils.queuePrefix + uuid;
		String rotation;
		String zentry = AnimatorUtils.getZ(nodeAttributes.get("zentry"), 0.3);
		double xentry = Double.parseDouble(nodeAttributes.get("xentry")), yentry = Double.parseDouble(nodeAttributes.get("yentry")), 
			xexit = Double.parseDouble(nodeAttributes.get("xexit")), yexit = Double.parseDouble(nodeAttributes.get("yexit"));	
		
		// get angle between entry and exit, adjust rotation accordingly
		double angle = AnimatorUtils.getAngle(xentry, xexit, yentry, yexit);
		// adjust position
		double xcenter = AnimatorUtils.getCenter(xentry, xexit);    		
		double ycenter  = AnimatorUtils.getCenter(yentry, yexit); 
		// get distance between entry and exit for length of queue
		String height = String.valueOf((AnimatorUtils.getDistance(xentry, xexit, yentry, yexit)));
		
		if ((angle == 0) || (angle == 180)) rotation = "0 0 1 1.57"; else rotation = Math.toRadians(angle)+" 0 0 1.57";
		scene.appendChild(ax3d.createBranchCylinder(defID, xcenter+" "+zentry+" -"+ycenter, rotation, height, ".15", ".6", ""));
		
		AnimatorObject ao = new AnimatorObject(defID, uuid, name, "queue");
		ao.setPosition(nodeAttributes.get("xentry"), nodeAttributes.get("yentry"), nodeAttributes.get("zentry"),
				nodeAttributes.get("xexit"), nodeAttributes.get("yexit"), nodeAttributes.get("zexit"));
		objects.put(uuid, ao);
	}
	
	/**
	 * Create a sink in x3d
	 */
	private void createObjectSink(String uuid, String name) {
		AnimatorObject sink = objects.get(uuid);
		String defID = AnimatorUtils.sinkPrefix + uuid;
		String x = nodeAttributes.get("xposition"), y = nodeAttributes.get("yposition"), 
			z = AnimatorUtils.getZ(nodeAttributes.get("zposition"), 0.5);
		
		// add label object
		if (AnimatorUtils.enableLabels)
			scene.appendChild(ax3d.createBranchText("", x+" "+String.valueOf(Double.parseDouble(z)+3)+" -"+y, "", name, "", "1 1 1"));
		
		// add sink object
		createTemplate(sink, "", defID, "SourceSink", x+" "+z+" -"+y, "");	
		
		// if sink not present, create it and set its position
		if (sink == null) {
			sink = new AnimatorObject(defID, uuid, name, "sink");
			objects.put(uuid, sink);
		}
		sink.setPosition(x, y, z);
	}
	
	/**
	 * Create a source in x3d
	 */
	private void createObjectSource(String uuid, String name) {
		AnimatorObject source = objects.get(uuid);
		String defID = AnimatorUtils.sourcePrefix + uuid;
		String x = nodeAttributes.get("xposition"), y = nodeAttributes.get("yposition"),  
			z = AnimatorUtils.getZ(nodeAttributes.get("zposition"), 0.5);
		
		// add label object
		if (AnimatorUtils.enableLabels)
			scene.appendChild(ax3d.createBranchText("", x+" "+String.valueOf(Double.parseDouble(z)+3)+" -"+y, "", name, "", "1 1 1"));
		
		// add source object
		createTemplate(source, "", defID, "SourceSink", x+" "+z+" -"+y, "");	
		
		// if source not present, create it and set its position
		if (source == null) {
			source = new AnimatorObject(defID, uuid, name, "source");
			objects.put(uuid, source);
		}
		source.setPosition(x, y, z);
	}
	
	/**
	 * Create a transporter in x3d
	 */
	private void createObjectTransporter(String uuid, String name) {
		AnimatorObject transporter = objects.get(uuid);
		String defID = AnimatorUtils.transporterPrefix + uuid;
		String x = nodeAttributes.get("xposition"), y = nodeAttributes.get("yposition"), z = nodeAttributes.get("zposition");
				
		// add transporter object
		createTemplate(transporter, defID+"trans", defID, "StandardTransporter", x+" "+z+" -"+y, "");			
		
		// if transporter not present, create it and set its position
		if (transporter == null) {
			transporter = new AnimatorObject(defID, uuid, name, "transporter");
			objects.put(uuid, transporter);
		}
		transporter.setPosition(x, y, z);
	}
	
	/**
	 * Create a workstation in x3d
	 */
	private void createObjectWorkstation(String uuid, String name) {
		AnimatorObject workstation = objects.get(uuid);
		String defID = AnimatorUtils.workstationPrefix + uuid;
		String x = nodeAttributes.get("xposition"), y = nodeAttributes.get("yposition"), z = nodeAttributes.get("zposition");
		
		// add label object
		if (AnimatorUtils.enableLabels)
			scene.appendChild(ax3d.createBranchText("", x+" "+String.valueOf(Double.parseDouble(z)+3)+" -"+y, "", name, "", "1 1 1"));
		
		// add workstation object
		createTemplate(workstation, "", defID, "StandardWorkstation", x+" "+z+" -"+y, "");
		
		// if workstation not present, create it and set its position
		if (workstation == null) {
			workstation = new AnimatorObject(defID, uuid, name, "workstation");
			objects.put(uuid, workstation);
		}
		workstation.setPosition(x, y, z);
	}
	
	/**
	 * create x3d object based on standard template/shape or custom template (proto/inline)
	 */
	private void createTemplate(AnimatorObject ao, String transDef, String defID, String stdTemplate, String position, String color) {
		boolean useCustom;
		if (ao == null) useCustom = false; else useCustom = !ao.getTemplateURL().isEmpty();

		if (useCustom) {
			// use proto instance if template is a prototype
			if (AnimatorUtils.isProtoDeclaration(ao.getTemplateURL()))
				scene.appendChild(ax3d.createBranchProtoInstance(transDef, defID, AnimatorUtils.getProtoName(ao.getTemplateURL()), position, ao.getRotationX3D(), color));
			// use inline (also check alternate locations) if template is no prototype
			else scene.appendChild(ax3d.createBranchInline(transDef, position, ao.getRotationX3D(), new String[]{
				ao.getTemplateURL(), 
				".\\templates\\"+new File(ao.getTemplateURL()).getName(),
				"..\\templates\\"+new File(ao.getTemplateURL()).getName(),
				"..\\..\\"+new File(ao.getTemplateURL()).getName(),
				"..\\"+new File(ao.getTemplateURL()).getName(),
				new File(ao.getTemplateURL()).getName()}));
		} else {
			// use standard template where standard template exists
			if (stdTemplate.substring(0, 8).equals("Standard"))
				scene.appendChild(ax3d.createBranchProtoInstance(transDef, defID, stdTemplate, position, "", color));
			// use standard source/sink
			if (stdTemplate.equals("SourceSink"))
				scene.appendChild(ax3d.createBranchBox(defID, position, "", ".5 .5 .5", "", ""));
		}
	}
	
	/**
	 * Add the animation
	 * Add optional baseplate, viewpoint position and finish the x3d document (create the physical file)
	 */
	public void createX3DFile() {
		// user interface - time sensor
		if (AnimatorUtils.useExtensiveMenu) {
			// add extensive user menu
			scene.appendChild(ax3d.createCustomMenuHUD(timesensorName, Integer.toString(AnimatorUtils.animationLength), String.valueOf((int)Math.ceil(Double.parseDouble(simulationLength)))));
			scene.appendChild(ax3d.createNodeProximitySensor("sensorMenuHUD", "", "1000000000 1000000000 1000000000"));
			scene.appendChild(ax3d.createNodeRoute("sensorMenuHUD", "position_changed", "transMenuHUD", "set_translation"));
			scene.appendChild(ax3d.createNodeRoute("sensorMenuHUD", "orientation_changed", "transMenuHUD", "set_rotation"));
		} else {
			// add normal time sensor
			scene.appendChild(ax3d.createNodeTimeSensor(timesensorName, Integer.toString(AnimatorUtils.animationLength), "", ""));
			scene.appendChild(ax3d.createNodeRoute(touchsensorName, "touchTime", timesensorName, "startTime"));
		}
		
		// add events
		createAnimationWorkstations();
		createAnimationJobs();
		if (AnimatorUtils.enableTransporters) createAnimationTransporters();

		// calculate viewpoint position and center of rotation
		double x = AnimatorUtils.getCenter(xmin, xmax);
		double y = (ymax-ymin)/2;
		double z = y/2;
		String cor = AnimatorUtils.getCenter(xmin, xmax)+" 0 -"+AnimatorUtils.getCenter(ymin, ymax);	
		createBaseplate(AnimatorUtils.showGrid, AnimatorUtils.showBaseplate, -0.05, -0.1, 5);	
		scene.appendChild(ax3d.createNodeViewpoint(x, y, z, cor));	
		scene.appendChild(ax3d.createNodeWorldInfo(AnimatorUtils.x3dFileName, ""));
		
		AnimatorXML.createXMLFile(ax3d.finishDocument(head, scene), AnimatorUtils.x3dFileName, "", "x3d", true);
	}
	
	/**
	 * Gets called whenever a closing tag is detected,
	 * used to close a branch and generate the object/event or gather further information
	 */
	public void finishBranch(String closingBranch) {
		// branch gets closed
		if (firstLevelBranch.equals(closingBranch)) {
			if (firstLevelBranch.equals("object") && AnimatorUtils.parseMode == AnimatorUtils.OPTIONS)
				setTemplateObjects();	// get object data for display in template options					
			if (firstLevelBranch.equals("object") && AnimatorUtils.parseMode != AnimatorUtils.OPTIONS) 
				addObject();	// write to x3d
			if (firstLevelBranch.equals("event") && AnimatorUtils.parseMode != AnimatorUtils.OPTIONS) 
				addEvent();		// write to x3d
			nodeAttributes.clear();
		}
		
		// branch still open, capture data of all nodes except empty nodes
		if ((firstLevelBranch.equals("object") || firstLevelBranch.equals("event")) && !currentTagValue.trim().isEmpty()) 
			setNodeAttributes(closingBranch);
		
		// set x-y-borders for the model (xmin/xmax and ymin/ymax), used for viewpoint calculation
		if (firstLevelBranch.equals("object") && AnimatorUtils.isNumber(currentTagValue)) 
			setXYBorders(closingBranch);
	}
	
	/**
	 * Identify the currently parsed branch on (first-level/second-level)
	 */
	public void identifyBranch(String node) {
		currentTagValue = "";
		// identify object node (structure and templates mode only)
		if (node.equals("object") && ((AnimatorUtils.parseMode == AnimatorUtils.X3D_STRUCT) ||
				(AnimatorUtils.parseMode == AnimatorUtils.OPTIONS))) firstLevelBranch = node;
		// identify event node (trace mode only)
		if (node.equals("event") && (AnimatorUtils.parseMode == AnimatorUtils.X3D_TRACE))
				firstLevelBranch = node;
		// identify second level node (all modes)
		if (node.equals("position") || node.equals("entry") || node.equals("exit") || 
				node.equals("from") || node.equals("to")) secondLevelBranch = node;
	}
	
	/**
	 * Save attributes of the currently parsed first-level branch to a hashmap
	 */
	private void setNodeAttributes(String nodeType) {
		// in case of x/y/z tags differentiate by adding the second-level-branch
		if (nodeType.length() == 1) nodeAttributes.put(nodeType+secondLevelBranch, currentTagValue);
		else nodeAttributes.put(nodeType, currentTagValue);
	}
	
	/**
	 * Return the current tag value
	 */
	public void setTagValue(String tagValue) {
		currentTagValue = tagValue;
	}
	
	/**
	 * Get all structure informations for options/templates
	 */
	private void setTemplateObjects() {
		String uuid = nodeAttributes.get("uuid");
		// set conveyors
		if (nodeAttributes.get("type").equals("conveyor")) {
			AnimatorObject ao = new AnimatorObject(AnimatorUtils.conveyorPrefix + uuid, uuid, nodeAttributes.get("name"), "conveyor");
			ao.setPosition(nodeAttributes.get("xfrom"), nodeAttributes.get("yfrom"), nodeAttributes.get("zfrom"), 
					nodeAttributes.get("xto"), nodeAttributes.get("yto"), nodeAttributes.get("zto"));
			ao.setStdTravelTime(nodeAttributes.get("stdtraveltime"));
			objects.put(uuid, ao);
		}
		// set parking
		if (nodeAttributes.get("type").equals("parking")) {
			AnimatorObject ao = new AnimatorObject(AnimatorUtils.parkingPrefix + uuid, uuid, nodeAttributes.get("name"), "parking");
			ao.setPosition(nodeAttributes.get("xposition"), nodeAttributes.get("yposition"), nodeAttributes.get("zposition"));
			objects.put(uuid, ao);
		}
		// set paths
		if (nodeAttributes.get("type").equals("path")) {
			AnimatorObject ao = new AnimatorObject(AnimatorUtils.pathPrefix + uuid, uuid, nodeAttributes.get("name"), "path");
			ao.setPosition(nodeAttributes.get("xfrom"), nodeAttributes.get("yfrom"), nodeAttributes.get("zfrom"), 
					nodeAttributes.get("xto"), nodeAttributes.get("yto"), nodeAttributes.get("zto"));
			ao.setStdTravelTime(nodeAttributes.get("stdtraveltime"));
			objects.put(uuid, ao);
		}
		// set queues
		if (nodeAttributes.get("type").equals("queue")) {
			AnimatorObject ao = new AnimatorObject(AnimatorUtils.queuePrefix + uuid, uuid, nodeAttributes.get("name"), "queue");
			ao.setPosition(nodeAttributes.get("xentry"), nodeAttributes.get("yentry"), nodeAttributes.get("zentry"), 
					nodeAttributes.get("xexit"), nodeAttributes.get("yexit"), nodeAttributes.get("zexit"));
			objects.put(uuid, ao);
		}
		// set items
		if (nodeAttributes.get("type").equals("item")) {
			AnimatorObject ao = new AnimatorObject(AnimatorUtils.itemPrefix + uuid, uuid, nodeAttributes.get("name"), "item");
			ao.setTemplateURL(nodeAttributes.get("template"));
			ao.setColor(AnimatorUtils.getRGBColorFromString(nodeAttributes.get("color")));
			objects.put(uuid, ao);
		}
		// set sinks
		if (nodeAttributes.get("type").equals("sink")) {
			AnimatorObject ao = new AnimatorObject(AnimatorUtils.sinkPrefix + uuid, uuid, nodeAttributes.get("name"), "sink");
			ao.setTemplateURL(nodeAttributes.get("template"));
			ao.setRotation(nodeAttributes.get("rotation"));
			ao.setPosition(nodeAttributes.get("xposition"), nodeAttributes.get("yposition"), nodeAttributes.get("zposition"));
			objects.put(uuid, ao);
		}
		// set sources
		if (nodeAttributes.get("type").equals("source")) {
			AnimatorObject ao = new AnimatorObject(AnimatorUtils.sourcePrefix + uuid, uuid, nodeAttributes.get("name"), "source");
			ao.setTemplateURL(nodeAttributes.get("template"));
			ao.setRotation(nodeAttributes.get("rotation"));
			ao.setPosition(nodeAttributes.get("xposition"), nodeAttributes.get("yposition"), nodeAttributes.get("zposition"));
			objects.put(uuid, ao);
		}
		// set transporters
		if (nodeAttributes.get("type").equals("transporter")) {
			AnimatorObject ao = new AnimatorObject(AnimatorUtils.transporterPrefix + uuid, uuid, nodeAttributes.get("name"), "transporter");
			ao.setTemplateURL(nodeAttributes.get("template"));
			ao.setPosition(nodeAttributes.get("xposition"), nodeAttributes.get("yposition"), nodeAttributes.get("zposition"));
			objects.put(uuid, ao);
		}
		// set workstations
		if (nodeAttributes.get("type").equals("workstation")) {
			AnimatorObject ao = new AnimatorObject(AnimatorUtils.workstationPrefix + uuid, uuid, nodeAttributes.get("name"), "workstation");
			ao.setTemplateURL(nodeAttributes.get("template"));
			ao.setRotation(nodeAttributes.get("rotation"));
			ao.setPosition(nodeAttributes.get("xposition"), nodeAttributes.get("yposition"), nodeAttributes.get("zposition"));
			objects.put(uuid, ao);
		}	
	}
	
	/**
	 * Set the min and max x/y borders (needed for viewpoint calculation)
	 */
	private void setXYBorders(String node) {
		double value = 0;
		if (!currentTagValue.isEmpty()) value = Double.parseDouble(currentTagValue);
		
		if (node.equals("x")) {
			if ((xmin == null) || (value < xmin)) xmin = value;
			if ((xmax == null) || (value > xmax)) xmax = value;
		}
		if (node.equals("y")) {
			if ((ymin == null) || (value < ymin)) ymin = value;
			if ((ymax == null) || (value > ymax)) ymax = value;
		}
	}
}
