/*
** Copyright (C) 2009 Patrick Kirchhof, Kevin Suske and Pasqual Dhring
** 
** This file is part of GFA.
** 
** GFA is free software: you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation, either version 3 of the License, or
** (at your option) any later version.
** 
** GFA is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with GFA.  If not, see <http://www.gnu.org/licenses/>.
** 
*/

package animator.src;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;

import java.io.File;

import javax.swing.BorderFactory;
import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.JButton;
import javax.swing.JColorChooser;
import javax.swing.JFileChooser;
import javax.swing.SpinnerListModel;
import javax.swing.filechooser.FileFilter;
import javax.swing.ImageIcon;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JSpinner;
import javax.swing.SpinnerModel;
import javax.swing.JTextField;

public class AnimatorFileBrowser extends JPanel {

	private static final long serialVersionUID = 6315764492599264621L;
	private Color c;
	private JButton button, bcolor; 
	private JLabel label;
	private JTextField text;
	private JSpinner rotation;
	private String ext, oldText;
	
	/**
	 * Creates new file browser panel which includes a label,
	 * a textfield and a button for browsing
	 */
	public AnimatorFileBrowser(String labelText, int spacing, String ext, boolean addColorChooser, boolean addOrientation) {
		button = new JButton("..."); 
		button.setToolTipText("Browse");
		label = new JLabel(labelText);
		text = new JTextField(30);
		if ((new File(AnimatorUtils.menuFolder+"ColorChooser.gif")).exists()) bcolor = new JButton("", new ImageIcon(AnimatorUtils.menuFolder+"ColorChooser.gif", ""));
			else if ((new File(AnimatorUtils.menuFolderRel+"ColorChooser.gif")).exists()) bcolor = new JButton("", new ImageIcon(AnimatorUtils.menuFolderRel+"ColorChooser.gif", ""));
		bcolor.setToolTipText("Color");
		
		// create new spinner with possible rotations
		String[] orientationArray = new String[361];
		for (int i = 0; i <= 360; i++) orientationArray[i] = i+"";
		rotation = new JSpinner(new SpinnerOrientationModel(orientationArray));
		rotation.setToolTipText("Rotation");
		rotation.setPreferredSize(new Dimension(50, 20));
		
		this.ext = ext;

		setLayout(new BoxLayout(this, BoxLayout.LINE_AXIS));
		setBorder(BorderFactory.createEmptyBorder(10,10,10,10));
		
		add(label); add(Box.createRigidArea(new Dimension(spacing, 0)));
		add(text); 
		if (addOrientation) add(rotation);
		if (addColorChooser) add(bcolor);
		add(button);		
		
		text.setEditable(false);
		button.setPreferredSize(new Dimension(20, 20));	
		button.addMouseListener(new MouseAdapter() {
            public void mouseClicked(MouseEvent e) {
            	oldText = text.getText();
            	text.setText(openFileChooser());
            	if ((label.getText().equals("Structure")) || (label.getText().equals("Trace")))
            		AnimatorFrontend.xmlLocationChanged = true;
            }
		});
		bcolor.setPreferredSize(new Dimension(20, 20));
		bcolor.addMouseListener(new MouseAdapter() {
            public void mouseClicked(MouseEvent e) {
            	c = JColorChooser.showDialog(null, "Color Chooser", c);  	
            }
		});	
	}
	
	/**
	 * Clear the text in the textfield
	 */
	public void clearText() {
		text.setText("");
	}
	
	/**
	 * Return the absolute path of the chosen file
	 */
	public String getAbsolutePath() {
		return text.getText();
	}
	
	/**
	 * Return the chosen color
	 */
	public Color getColor() {
		return c;
	}
	
	/**
	 * Return the extension of a file
	 */
	public static String getExtension(File f) {
        String ext = null;
        String s = f.getName();
        int i = s.lastIndexOf('.');

        if (i > 0 &&  i < s.length() - 1) {
            ext = s.substring(i+1).toLowerCase();
        }
        return ext;
    }
	
	/**
	 * Return the chosen rotation
	 */
	public String getRotation() {
		return (String)rotation.getValue();
	}
	
	/**
	 * Open a window for file browsing; file filtering by 
	 * extension is also done here
	 */
	public String openFileChooser() {
		JFileChooser chooser = new JFileChooser();

		if (ext.equals("xml")) {
			chooser.setCurrentDirectory(new File("."));
			chooser.addChoosableFileFilter(new OneTypeFilter("xml"));
		}
		if (ext.equals("x3d")) {
			if ((new File(AnimatorUtils.templateFolder)).exists()) chooser.setCurrentDirectory(new File(AnimatorUtils.templateFolder));
				else if ((new File(AnimatorUtils.templateFolderRel)).exists()) chooser.setCurrentDirectory(new File(AnimatorUtils.templateFolderRel));
			chooser.addChoosableFileFilter(new OneTypeFilter("x3d"));
		} 
		
	    int returnVal = chooser.showOpenDialog(this);
	    if(returnVal == JFileChooser.APPROVE_OPTION) return chooser.getSelectedFile().getAbsolutePath();
	    if(returnVal == JFileChooser.CANCEL_OPTION)  return oldText;
	    return "";
	}
	
	/**
	 * Revert to the text that was in the text field
	 * before changes took place
	 */
	public void revertText() {
		text.setText(oldText);
	}
	
	/**
	 * Set the color
	 */
	public void setColor(Color color) {
		this.c = color;
	}
	
	/**
	 * Set the rotation
	 */
	public void setRotation(String rotation) {
		this.rotation.setValue(rotation);
	}
	
	/**
	 * Set a text
	 */
	public void setText(String s) {
		text.setText(s);
	}
	
	/**
	 * Filter class which does all the filtering (by extension)
	 */
	private class OneTypeFilter extends FileFilter {
		private String type;
		
		public OneTypeFilter(String type) {
			this.type =type;
		}
		
		public boolean accept(File f) {
	        if (f.isDirectory()) {
	            return true;
	        }

	        String extension = getExtension(f);
	        if (extension != null) {
	            if (extension.equals(type)) return true;
	            	else return false;
	        }
	        return false;
	    }

		public String getDescription() {
			return type.toUpperCase()+"-Files";
		}
	}
	
	/**
	 * Spinner class which accepts an array for its values
	 */
	private class SpinnerOrientationModel extends SpinnerListModel {
		private static final long serialVersionUID = 9081950293405602895L;
		Object firstValue, lastValue;
	    SpinnerModel linkedModel = null;

	    public SpinnerOrientationModel(Object[] values) {
	        super(values);
	        firstValue = values[0];
	        lastValue = values[values.length - 1];
	    }

	    public void setLinkedModel(SpinnerModel linkedModel) {
	        this.linkedModel = linkedModel;
	    }

	    public Object getNextValue() {
	        Object value = super.getNextValue();
	        if (value == null) {
	            value = firstValue;
	            if (linkedModel != null) {
	                linkedModel.setValue(linkedModel.getNextValue());
	            }
	        }
	        return value;
	    }

	    public Object getPreviousValue() {
	        Object value = super.getPreviousValue();
	        if (value == null) {
	            value = lastValue;
	            if (linkedModel != null) {
	                linkedModel.setValue(linkedModel.getPreviousValue());
	            }
	        }
	        return value;
	    }
	}

}
