/*
** Copyright (C) 2009 Patrick Kirchhof, Kevin Suske and Pasqual Dhring
** 
** This file is part of GFA.
** 
** GFA is free software: you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation, either version 3 of the License, or
** (at your option) any later version.
** 
** GFA is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with GFA.  If not, see <http://www.gnu.org/licenses/>.
** 
*/

package animator.src;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

public class AnimatorX3D {
	
	private Document x3dDoc;
	
	/**
	 * This class offers methods which create x3d nodes and branches according to x3d conventions
	 */
	public AnimatorX3D() {
		x3dDoc = AnimatorXML.createXMLDocument();
	}
	
	public Element createBranchBox(String shapeDef, String translationXZY, String rotation, 
			String sizeXYZ, String transparency, String diffuseColorRGB) {
		Element transform = createNodeTransform("", translationXZY, rotation);
		Element shape = createNodeShape(shapeDef);
		Element appearance = createNodeAppearance();
		
		// level 3									// solid objects: use diffuseColor for light reflections
		appearance.appendChild(createNodeMaterial("", transparency, diffuseColorRGB, ""));
		// level 2
		shape.appendChild(appearance);
		shape.appendChild(createNodeBox(sizeXYZ));
		// level 1
		transform.appendChild(shape);
		
		return transform;
	}
	
	public Element createBranchCylinder(String shapeDef, String translationXZY, String rotation, 
			String height, String radius, String transparency, String diffuseColorRGB) {
		Element transform = createNodeTransform("", translationXZY, rotation);
		Element shape = createNodeShape(shapeDef);
		Element appearance = createNodeAppearance();
		
		// level 3									// solid objects: use diffuseColor for light reflections
		appearance.appendChild(createNodeMaterial("", transparency, diffuseColorRGB, ""));
		// level 2
		shape.appendChild(appearance);
		shape.appendChild(createNodeCylinder(height, radius));
		// level 1
		transform.appendChild(shape);
		
		return transform;
	}
	
	public Element createBranchInline(String transDef, String translationXZY, String rotation, String[] urls) {
		Element transform = createNodeTransform(transDef, translationXZY, rotation);
		
		StringBuffer url = new StringBuffer();
		for (int i = 0; i < urls.length; i++) url.append("\""+urls[i]+"\"");
		transform.appendChild(createNodeInline("", url.toString()));
		
		return transform;
	}
	
	public Element createBranchLineSet(String shapeDef, String vertexCount, String point, String transparency, 
			String emissiveColorRGB, String linewidthScaleFactor) {
		Element transform = createNodeTransform("", "", "");
		Element shape = createNodeShape(shapeDef);
		Element appearance = createNodeAppearance();
		Element lineset = createNodeLineSet(vertexCount, "geometry");
		
		// level 3									// line objects: use emissiveColor only
		appearance.appendChild(createNodeMaterial("", transparency, "", emissiveColorRGB));
		appearance.appendChild(createNodeLineProperties("true", "1", linewidthScaleFactor));
		lineset.appendChild(createNodeCoordinate("", point));
		// level 2
		shape.appendChild(appearance);
		shape.appendChild(lineset);
		// level 1
		transform.appendChild(shape);
		
		return transform;
	}
	
	public Element createBranchProtoInstance(String transDef, String protoDef, String name, String translationXZY, String rotation, String color) {
		Element transform = createNodeTransform(transDef, translationXZY, rotation);
		Element proto = createNodeProtoInstance(name, protoDef);
		
		if (!color.isEmpty()) proto.appendChild(createNodeFieldValue("color", color));
		transform.appendChild(proto);
		
		return transform;
	}
	
	public Element createBranchQuadSet(String shapeDef, String point, String transparency, String emissiveColorRGB) {
		Element transform = createNodeTransform("", "", "");
		Element shape = createNodeShape(shapeDef);
		Element appearance = createNodeAppearance();
		Element lineset = createNodeQuadSet("false", "true", "true", "true", "geometry");
		
		// level 3
		appearance.appendChild(createNodeMaterial("", transparency, "", emissiveColorRGB));
		lineset.appendChild(createNodeCoordinate("", point));
		// level 2
		shape.appendChild(appearance);
		shape.appendChild(lineset);
		// level 1
		transform.appendChild(shape);
		
		return transform;
	}
	
	public Element createBranchText(String shapeDef, String translationXZY, String rotation, 
			String string, String transparency, String diffuseColorRGB) {
		Element transform = createNodeTransform("", translationXZY, rotation);
		Element billboard = createNodeBillboard("0 0 0");
		Element shape = createNodeShape(shapeDef);
		Element appearance = createNodeAppearance();
		Element text = createNodeText(string, "true");
		
		// level 4									// solid objects: use diffuseColor for light reflections
		appearance.appendChild(createNodeMaterial("", transparency, diffuseColorRGB, ""));
		text.appendChild(createNodeFontStyle("SANS", "PLAIN", "MIDDLE", "0.5"));
		// level 3
		shape.appendChild(appearance);
		shape.appendChild(text);
		// level 2
		billboard.appendChild(shape);
		// level 1
		transform.appendChild(billboard);
		
		return transform;
	}
	
	public Element createCustomGroupUI(String touchsensorName) {
		Element group = createNodeGroup();
		Element transform_outer = createNodeTransform("UserInterface", "0 0 0", "0 0 1 0");
		Element transform_inner = createNodeTransform("", "-1 1 -3", "");
		Element shape = createNodeShape("");
		Element appearance = createNodeAppearance();
		
		// level 4
		appearance.appendChild(createNodeMaterial("", "", "1 0 0", ""));
		// level 3
		shape.appendChild(appearance);
		shape.appendChild(createNodeBox(".2 .2 .01"));
		// level 2
		transform_inner.appendChild(shape);
		transform_inner.appendChild(createNodeTouchSensor(touchsensorName));
		transform_outer.appendChild(transform_inner);
		// level 1
		group.appendChild(createNodeProximitySensor("ProximitySensorOrigin", "0 0 0", "10000 10000 10000"));
		group.appendChild(transform_outer);
		
		return group;
	}
	
	public Element createCustomMenuHUD(String timeDef, String cycleInterval, String max_time) {
		Element transform_outer = createNodeTransform("transMenuHUD", "", "");
		Element transform_inner = createNodeTransform("", "-1.5 -3.0 -10.0", "");
		Element proto = createNodeProtoInstance("TimeMenu", timeDef);
		
		// level 3
		proto.appendChild(createNodeFieldValue("cycleInterval", cycleInterval));
		proto.appendChild(createNodeFieldValue("max_time", max_time));
		// level 2
		transform_inner.appendChild(proto);
		// level 1
		transform_outer.appendChild(transform_inner);
		
		return transform_outer;
	}
	
	public Element createNodeAppearance() {
		Element e = x3dDoc.createElement("Appearance");
		return e;
	}
	
	public Element createNodeBillboard(String axisOfRotation) {
		Element e = x3dDoc.createElement("Billboard");
		e.setAttribute("axisOfRotation", axisOfRotation);
		return e;
	}
	
	public Element createNodeBooleanSequencer(String def, String key, String keyValue) {
		Element e = x3dDoc.createElement("BooleanSequencer");
		e.setAttribute("DEF", def);
		e.setAttribute("key", key);
		e.setAttribute("keyValue", keyValue);
		return e;
	}
	
	public Element createNodeBox(String sizeXYZ) {
		Element e = x3dDoc.createElement("Box");
		e.setAttribute("size", sizeXYZ);
		return e;
	}
	
	public Element createNodeCoordinate(String def, String point) {
		Element e = x3dDoc.createElement("Coordinate");
		if (!def.isEmpty()) e.setAttribute("DEF", def);
		e.setAttribute("point", point);
		return e;
	}
	
	public Element createNodeCylinder(String height, String radius) {
		Element e = x3dDoc.createElement("Cylinder");
		e.setAttribute("height", height);
		e.setAttribute("radius", radius);
		return e;
	}
	
	public Element createNodeExternProtoDeclare(String name, String[] urls) {
		Element e = x3dDoc.createElement("ExternProtoDeclare");
		e.setAttribute("name", name);
		
		StringBuffer url = new StringBuffer();
		for (int i = 0; i < urls.length; i++) url.append("\""+urls[i]+"\"");
		e.setAttribute("url", url.toString());
		return e;
	}
	
	public Element createNodeField(String accessType, String name, String type) {
		Element e = x3dDoc.createElement("field");
		e.setAttribute("accessType", accessType);
		e.setAttribute("name", name);
		e.setAttribute("type", type);
		return e;
	}
	
	public Element createNodeFieldValue(String name, String value) {
		Element e = x3dDoc.createElement("fieldValue");
		e.setAttribute("name", name);
		e.setAttribute("value", value);
		return e;
	}
	
	public Element createNodeFontStyle(String family, String style, String justify, String size) {
		Element e = x3dDoc.createElement("FontStyle");
		e.setAttribute("family", family);
		e.setAttribute("style", style);
		e.setAttribute("justify", justify);
		e.setAttribute("size", size);
		return e;
	}
	
	public Element createNodeGroup() {
		Element e = x3dDoc.createElement("Group");
		return e;
	}
	
	public Element createNodeHead() {
		Element e = x3dDoc.createElement("head");
		return e;
	}
	
	public Element createNodeInline(String def, String url) {
		Element e = x3dDoc.createElement("Inline");
		if (!def.isEmpty()) e.setAttribute("DEF", def);
		e.setAttribute("url", url);
		return e;
	}
	
	public Element createNodeIntegerSequencer(String def, String key, String keyValue) {
		Element e = x3dDoc.createElement("IntegerSequencer");
		e.setAttribute("DEF", def);
		e.setAttribute("key", key);
		e.setAttribute("keyValue", keyValue);
		return e;
	}
	
	public Element createNodeLineProperties(String applied, String linetype, String linewidthScaleFactor) {
		Element e = x3dDoc.createElement("LineProperties");
		e.setAttribute("applied", applied);
		e.setAttribute("linetype", linetype);
		e.setAttribute("linewidthScaleFactor", linewidthScaleFactor);
		return e;
	}
	
	public Element createNodeLineSet(String vertexCount, String containerField) {
		Element e = x3dDoc.createElement("LineSet");
		e.setAttribute("vertexCount", vertexCount);
		e.setAttribute("containerField", containerField);
		return e;
	}
	
	public Element createNodeMaterial(String def, String transparency, String diffuseColorRGB, String emissiveColorRGB) {
		Element e = x3dDoc.createElement("Material");
		if (!def.isEmpty()) e.setAttribute("DEF", def);
		if (!transparency.isEmpty()) e.setAttribute("transparency", transparency);
		if (!diffuseColorRGB.isEmpty()) e.setAttribute("diffuseColor", diffuseColorRGB);
		if (!emissiveColorRGB.isEmpty()) e.setAttribute("emissiveColor", emissiveColorRGB);
		return e;
	}
	
	public Element createNodeMeta(String name, String content) {
		Element e = x3dDoc.createElement("meta");
		e.setAttribute("name", name);
		e.setAttribute("content", content);
		return e;		
	}
	
	public Element createNodeOrientationInterpolator(String def, String key, String keyValue) {
		Element e = x3dDoc.createElement("OrientationInterpolator");
		e.setAttribute("DEF", def);
		e.setAttribute("key", key);
		e.setAttribute("keyValue", keyValue);
		return e;	
	}
	
	public Element createNodePositionInterpolator(String def, String key, String keyValue) {
		Element e = x3dDoc.createElement("PositionInterpolator");
		e.setAttribute("DEF", def);
		e.setAttribute("key", key);
		e.setAttribute("keyValue", keyValue);
		return e;	
	}
	
	public Element createNodeProtoInstance(String name, String def) {
		Element e = x3dDoc.createElement("ProtoInstance");
		e.setAttribute("DEF", def);
		e.setAttribute("name", name);
		return e;	
	}
	
	public Element createNodeProximitySensor(String def, String center, String size) {
		Element e = x3dDoc.createElement("ProximitySensor");
		e.setAttribute("DEF", def);
		if (!center.isEmpty()) e.setAttribute("center", center);
		e.setAttribute("size", size);
		return e;		
	}
	
	public Element createNodeQuadSet(String solid, String ccw, String colorPerVertex, String normalPerVertex, String containerField) {
		Element e = x3dDoc.createElement("QuadSet");
		e.setAttribute("solid", solid);
		e.setAttribute("ccw", ccw);
		e.setAttribute("colorPerVertex", colorPerVertex);
		e.setAttribute("normalPerVertex", normalPerVertex);
		e.setAttribute("containerField", containerField);
		return e;	
	}
	
	public Element createNodeRoute(String fromNode, String fromField, String toNode, String toField) {
		Element e = x3dDoc.createElement("ROUTE");
		e.setAttribute("fromNode", fromNode);
		e.setAttribute("fromField", fromField);
		e.setAttribute("toNode", toNode);
		e.setAttribute("toField", toField);
		return e;
	}
	
	public Element createNodeScene() {
		Element e = x3dDoc.createElement("Scene");
		return e;
	}
	
	public Element createNodeScript(String def, String directOutput, String mustEvaluate) {
		Element e = x3dDoc.createElement("Script");
		e.setAttribute("DEF", def);
		e.setAttribute("directOutput", directOutput);
		e.setAttribute("mustEvaluate", mustEvaluate);
		return e;
	}
	
	public Element createNodeShape(String def) {
		Element e = x3dDoc.createElement("Shape");
		if (!def.isEmpty()) e.setAttribute("DEF", def);
		return e;
	}
	
	public Element createNodeSphere(String radius) {
		Element e = x3dDoc.createElement("Sphere");
		e.setAttribute("radius", radius);
		return e;
	}
	
	public Element createNodeText(String string, String solid) {
		Element e = x3dDoc.createElement("Text");
		e.setAttribute("string", string);
		e.setAttribute("solid", solid);
		return e;
	}
	
	public Element createNodeTimeSensor(String def, String cycleInterval, String enabled, String loop) {
		Element e = x3dDoc.createElement("TimeSensor");
		e.setAttribute("DEF", def);
		e.setAttribute("cycleInterval", cycleInterval);
		if (!enabled.isEmpty()) e.setAttribute("enabled", enabled);
		if (!loop.isEmpty()) e.setAttribute("loop", loop);
		return e;
	}
	
	public Element createNodeTouchSensor(String def) {
		Element e = x3dDoc.createElement("TouchSensor");
		e.setAttribute("DEF", def);
		return e;
	}
	
	public Element createNodeTransform(String def, String translationXZY, String rotation) {
		Element e = x3dDoc.createElement("Transform");
		if (!def.isEmpty()) e.setAttribute("DEF", def);
		if (!translationXZY.isEmpty()) e.setAttribute("translation", translationXZY);
		if (!rotation.isEmpty()) e.setAttribute("rotation", rotation);
		return e;
	}
	
	public Element createNodeViewpoint(double xpos, double ypos, double zpos, String cor) {
		Element e = x3dDoc.createElement("Viewpoint");
		e.setAttribute("description", "Default View");		 
		e.setAttribute("orientation", "-1 0 0 0.5");
		e.setAttribute("position", xpos+" "+ypos+" "+zpos);
		e.setAttribute("centerOfRotation", cor);
		return e;
	}
	
	public Element createNodeWorldInfo(String title, String info) {
		Element e = x3dDoc.createElement("WorldInfo");
		e.setAttribute("title", title);		 
		if (!info.isEmpty()) e.setAttribute("info", info);
		return e;
	}
	
	public Element createNodeX3D() {
		Element e = x3dDoc.createElement("X3D");
		e.setAttribute("profile", "Immersive");
		e.setAttribute("version", "3.2");
		e.setAttribute("xmlns:xsd", "http://www.w3.org/2001/XMLSchema-instance");
		e.setAttribute("xsd:noNamespaceSchemaLocation", "http://www.web3d.org/specifications/x3d-3.2.xsd");
		return e;
	}
	
	public Document finishDocument(Element head, Element scene) {
		Element e = createNodeX3D();
		e.appendChild(head);
		e.appendChild(scene);
		x3dDoc.appendChild(e);
		return x3dDoc;
	}
	
	public Document getDocument() {
		return x3dDoc;
	}
}